import gameInterface.AIInterface;
import commandcenter.CommandCenter;
import structs.GameData;
import structs.FrameData;
import structs.Key;

import java.util.ArrayList;
import java.util.Random;

/**
 * Main class of ZoneAI for FightingICE
 * 
 * Inspired by the simplicity of the 2016 KeepYourDistanceBot and other rule-based algorithms
 * Currently only has simple rules without careful consideration of hitboxes and enemy range
 * 
 * @author Frank Ying
 *
 */

// TODO Different rules for close, mid, far range, character aerial state, time remaining, opponent energy...etc.
// TODO Fix LUD moves
// TODO Predict enemy attacks from recently used attacks

public class ZoneAI implements AIInterface 
{
	private Key inputKey;
	private CommandCenter commandCenter;
	private FrameData frameData;
	private GameData gameData;
	private Random rnd;

	private boolean playerNumber;
	private String myCharacter;
	
	private final int energyCost = 30;
	private final int trappedDistance = 200;
	private final int closeDistance = 175;
	private int farDistance = 500;
	
	private final int LEFT_WALL = -120;
	private final int RIGHT_WALL = 680;
	
	private String fireballKey;
	private String wallEscapeKey;
	private String escapeKey;
	private String gapCloserKey;
	private ArrayList<String> attackChoices;

	/**
	 * FIXED, DO NOT CHANGE
	 */
	@Override
	public String getCharacter() 
	{
		return CHARACTER_ZEN;
	}

	/**
	 * Initializes AI (runs only once at the beginning of the game)
	 * @param gd GameData structure
	 * @param player Player side; true for P1, false for P2
	 */
	@Override
	public int initialize(GameData gd, boolean player) 
	{
		this.inputKey = new Key();
		this.commandCenter = new CommandCenter();
		this.frameData = new FrameData();
		this.gameData = gd;
		this.playerNumber = player;
		this.rnd = new Random();
		this.attackChoices = new ArrayList<String>();

		// Get character name
		if (playerNumber)
			this.myCharacter = gameData.getPlayerOneCharacterName();
		else
			this.myCharacter = gameData.getPlayerTwoCharacterName();

		if (myCharacter == "ZEN")
		{
			attackChoices.add("STAND_B");
			attackChoices.add("CROUCH_A");
			attackChoices.add("CROUCH_B");
			attackChoices.add("CROUCH_FB");
			attackChoices.add("FOR_JUMP B");
			attackChoices.add("FOR_JUMP FA");
			attackChoices.add("FOR_JUMP FB");
			attackChoices.add("STAND_GUARD");
			this.gapCloserKey = "FOR_JUMP B";
			this.fireballKey = "STAND_D_DF_FB";
			this.wallEscapeKey = "FOR_JUMP";
			this.escapeKey = "BACK_JUMP";
			this.farDistance = 500;
		}

		else if (myCharacter == "GARNET")
		{
			attackChoices.add("STAND_A");
			attackChoices.add("STAND_FA");
			attackChoices.add("STAND_FB");
			attackChoices.add("CROUCH_A");
			attackChoices.add("FOR_JUMP FB");
			attackChoices.add("FOR_JUMP DA");
			attackChoices.add("FOR_JUMP DB");
			attackChoices.add("STAND_GUARD");
			this.gapCloserKey = "FOR_JUMP B";
			this.fireballKey = "STAND_D_DF_FA";
			this.wallEscapeKey = "FOR_JUMP";
			this.escapeKey = "BACK_JUMP";
			this.farDistance = 500;
		}
		else if (myCharacter == "LUD") 
		{
			attackChoices.add("STAND_A");
			attackChoices.add("STAND_FA");
			attackChoices.add("CROUCH_A");
			attackChoices.add("CROUCH_FA");
			attackChoices.add("AIR_A");
			attackChoices.add("AIR_FA");
			attackChoices.add("AIR_B");
			attackChoices.add("STAND_GUARD");
			this.gapCloserKey = "DASH";
			this.fireballKey = "STAND_D_DF_FA";
			this.wallEscapeKey = "AIR_DB";
			this.escapeKey = "AIR_DB";
			this.farDistance = 400;
		}

		return 0;
	}

	/**
	 * Gets information of current frame (runs every frame)
	 * @param fd FrameData structure
	 */
	@Override
	public void getInformation(FrameData fd) 
	{
		this.frameData = fd;
		commandCenter.setFrameData(frameData, playerNumber);
	}

	/**
	 * Main processing part (runs every frame)
	 */
	@Override
	public void processing() 
	{
		// Make sure frameData is not empty & round has started, to prevent NullPointerException
		if(!frameData.getEmptyFlag() && frameData.getRemainingTimeMilliseconds() > 0)
		{	
			if (commandCenter.getSkillFlag())
				inputKey = commandCenter.getSkillKey();
			else
			{
				inputKey.empty();
				commandCenter.skillCancel();

				// Calculate distance between characters and action according to the distance
				int distance = commandCenter.getDistanceX();
				if (distance <= closeDistance)
					escape(commandCenter, gameData);
				else if (distance > closeDistance && distance <= farDistance) 
					// attack
					attack(commandCenter);
				else
					commandCenter.commandCall(gapCloserKey);
			}
		}
	}

	/**
	 * Sends the input key from AI to game (runs every frame)
	 */
	@Override
	public Key input() 
	{
		return inputKey;
	}

	/**
	 * Finalizes AI (runs only once at the end of the game)
	 */
	@Override
	public void close() 
	{
	}

	/*
	 * 
	 */

	/**
	 * Method to determine which attack to use
	 * Uses fireball if it has enough energy
	 * Otherwise pick an attack with a good hitbox
	 * @param cc CommandCenter class
	 */
	private void attack(CommandCenter cc)
	{
		if (cc.getMyEnergy() >= energyCost)
			cc.commandCall(fireballKey);
		else
			cc.commandCall(attackChoices.get(rnd.nextInt(attackChoices.size())));
	}

	/**
	 * Method to determine which way to escape
	 * Escapes backwards under normal circumstances
	 * Escapes forwards when there is no more room behind
	 * Exception for LUD, who needs energy to jump forwards & backwards
	 * @param cc CommandCenter class
	 * @param gd GameData structure
	 */
	private void escape(CommandCenter cc, GameData gd)
	{
		if ((cc.getMyX() <= LEFT_WALL || cc.getMyX() >= RIGHT_WALL) && cc.getDistanceX() <= trappedDistance)
			cc.commandCall(wallEscapeKey);
		else
			cc.commandCall(escapeKey);
	}
}


